;===============================================
; Program:     Grid2DManager
; Version:     Beta 1.0.
; Date:        17.03.2008
; Author:      Michael 'neotoma' Taupitz
; TargetOS:    ALL.
; Compiler:    PureBasic 4.1 for Windows.
;
; Licence:     Do as you wish with it! No warranties either implied
;              or given... blah blah! :)
;
; Changes / History
; 2.4.2008 - Add Procedure to Free the GridManager2D
; 4.4.2008 - fixed existing Memory-Leaks
;===============================================

;Here comes the Grid2D-Manager
;
; This Manager creates and observes the Grid. 
; You can register objects to that Grid (so it will be recognized on Zones)
; And with that, you can get all objects that are in the same ziones as a
; given object !!
;
; Here goes the same thanks as for the GridObject2d (flype, Kaeru Gamen ect.)
;



;------- Includes ---
IncludeFile "_basic_macros.pbi"
IncludeFile "LinkedListAndTree2.pbi"
IncludeFile "GridObject2d.pbi"

;------- Constants ---
;


;------- Macros ---

;these Macros are intened to stored in a own include-file !

;Used not as expression. anly chganges the value of the lasr parameter
; to true or false. 
Macro mcCollissionStd(__x1__,__y1__,__a1__,__b1__,__x2__,__y2__,__a2__,__b2__,__V__) 
  If __x1__ < __x2__ : ___dx = __x2__ - (__x1__ + __a1__) : Else : ___dx = __x1__ - (__x2__ + __a2__) : EndIf
  If __y1__ < __y2__ : ___dy = __y2__ - (__y1__ + __b1__) : Else : ___dy = __y1__ - (__y2__ + __b2__) : EndIf 
  If ___dx < 0 And ___dy < 0 : __V__ = #True : Else : __V__ = #False  : EndIf
EndMacro  
  

Macro mcRectCollWrapper(__RECT1__,__RECT2__,__V__)
  
  If __RECT1__\X < __RECT2__\X : ___dx = __RECT2__\X - (__RECT1__\X + __RECT1__\width) : Else : ___dx = __RECT1__\X - (__RECT2__\X + __RECT2__\width) : EndIf
  If __RECT1__\Y < __RECT2__\Y : ___dy = __RECT2__\Y - (__RECT1__\Y + __RECT1__\width) : Else : ___dy = __RECT1__\Y - (__RECT2__\Y + __RECT2__\width) : EndIf
  If ___dx < 0 And ___dy < 0 : __V__ = #True : Else : __V__ = #False  : EndIf
   
  ;  mcCollissionStd(__RECT1__\x,__RECT1__\y,__RECT1__\width,__RECT1__\height , __RECT2__\x,__RECT2__\y,__RECT2__\width,__RECT2__\height)
EndMacro

Macro mcDebugRect(__RECT__)
  Debug("Rect("+Str(__RECT__\X)+", " +Str(__RECT__\Y)+", "+Str(__RECT__\width)+", "+Str(__RECT__\height)+")")
EndMacro

Macro mcDrawRect(__RECT__,__COLOR__=#Yellow)
  Box(__RECT__\X,__RECT__\Y,__RECT__\width,__RECT__\height,__COLOR__)
EndMacro

;------- Structures ---

; -----------------------------------------------------------------------------
; -  This Structure holds the information of a zone in the Grid.
; -----------------------------------------------------------------------------
Structure sGridObjects 
  rect.sRectL   ;recangle = bound of a Zone
  llGridObj.l   ;Linked List of Object in that Zone
EndStructure

; -----------------------------------------------------------------------------
; -  This Structure holds the information of a zone in the Object.
; -----------------------------------------------------------------------------
Structure sZonePtr Extends LL
  zone.l
EndStructure

; -----------------------------------------------------------------------------
; - This Structure hods a Pointer to a sGridObject
; -----------------------------------------------------------------------------
Structure sGridObjPtr Extends LL
  obj.l         ; pointer to a GridObject2D
  llZones.l     ; ll of affected Zones. (sZonePtr)
EndStructure

; -----------------------------------------------------------------------------
; - This Structure is the Main-Information about the Grid.
; - there are some fileds, that are unused at the moment
; -----------------------------------------------------------------------------
Structure sGrid
  name.s                  ;not important, but maybe for Debug-Relations
  
  id.l                    ;same here - but it should based on a 'unique' number.
  
  dimension.SIZE          ;Grid-Dimension
  gridElementSize.SIZE    ;Size of a Zone ;Needed to create the 
  
  llRegisteredObjects.l   ;Linked List of registerd Objects - needed for fas updating
  
  paZones.l               ;fixed Memory-Array for the Zones (each filled with a sGridObject !)
EndStructure


; -----------------------------------------------------------------------------
; This Procedure is used internally to create a Grid.
; Basicly it allocates the Memory for the Zones and initializes the Zones
; with the Rect-values and also creating the LinkedList for the Objects in
; each Zone.

Procedure _internal_buildGrid(*grid.sGrid)
  
  *grid\paZones = AllocateMemory(  *grid\dimension\cx * *grid\dimension\cy * SizeOf(sGridObjects))
  
  ;Do some precalc. btw: this could be also stored in the 'sGrid', so we need no
  ;not to recalculate this later.
  sz = SizeOf(sGridObjects)
  wz = *grid\dimension\cx*sz
  
  For X = 0 To *grid\dimension\cx-1
    For Y = 0 To *grid\dimension\cy-1
      
      *zone.sGridObjects = *grid\paZones + (( Y*wz)+(X*sz))
      *zone\rect\X = X* *grid\gridElementSize\cx
      *zone\rect\Y = Y* *grid\gridElementSize\cy 
      *zone\rect\width = *grid\gridElementSize\cx
      *zone\rect\height = *grid\gridElementSize\cy 
      *zone\llGridObj = NewList2(sGridObjPtr) 
      mcDebugRect(*zone\rect)
       
    Next
  Next
  
EndProcedure


; -----------------------------------------------------------------------------
; - lPointer =  GridManager2D_Create( dimx.l, dimy.l, elsizex.l, elsizey.l, name.s="unknown", id.l=-1)
; -----------------------------------------------------------------------------
; - Description:
; - With this Procedur you create aGrind in the Dimension dimx,dimy. 
; - and each zone in this grid has asSize of elsizex,elsizey.
; - for example - a Screen of 800x600 with Zone size 32x32 would be
; -
; - #GRIDZONE=32
; - *grid = GridManager2D_Create(800/#GRIDZONE, 600/#GRIDZONE,#GRIDZONE,#GRIDZONE,"mygrid")
; - 
; -----------------------------------------------------------------------------
; - Returns:
; - Pointer to a valid Grid (sGrid-Structure) if succeeds - otherwise #NULL
; -----------------------------------------------------------------------------

Procedure.l GridManager2D_Create( dimx.l, dimy.l, elsizex.l, elsizey.l, name.s="unknown", id.l=-1)
  *mem.sGrid = AllocateMemory(SizeOf(sGrid))
  If *mem
    *mem\name = name
    *mem\id = id
    *mem\dimension\cx = dimx
    *mem\dimension\cy = dimy
    *mem\gridElementSize\cx = elsizex 
    *mem\gridElementSize\cy = elsizey
    *mem\llRegisteredObjects = NewList2(sGridObjPtr)
    _internal_buildGrid(*mem)
  Else
    Debug("GridManager2D_Create("+name+"): Can't get Memory for Structure : sGrid" )
  EndIf 
  ProcedureReturn *mem
EndProcedure  


; -----------------------------------------------------------------------------
; - lPointer =  GridManager2D_Free( *gm.sGrid)
; -----------------------------------------------------------------------------
; - Description:
; - Here we Destroy the Grid with all Linked lists and make everything
; - nice and clesn.  The Objects can't be destroyed, there is no 'Default'
; - for destroying them. So this part is up to the programm who added the
; - Objects.
; - 
; -----------------------------------------------------------------------------
; - Returns:
; - Nothing
; -----------------------------------------------------------------------------

Procedure.l GridManager2D_Free( *gm.sGrid) 
  If *gm 
    *reg_obj.sGridObjPtr = FirstElement2(*gm\llRegisteredObjects) 
    While *reg_obj  
       DeleteList2(*reg_obj\llZones) 
      *reg_obj = NextElement2(*reg_obj) 
    Wend   
    DeleteList2(*gm\llRegisteredObjects)
    
    sz = SizeOf(sGridObjects)
    wz = *gm\dimension\cx*sz

    For X = 0 To *gm\dimension\cx-1
      For Y = 0 To *gm\dimension\cy-1
        *zone.sGridObjects = *gm\paZones + (( Y*wz)+(X*sz))
        DeleteList2(*zone\llGridObj)
      Next
    Next
    FreeMemory(*gm\paZones)
    FreeMemory(*gm)
    
  Else
    Debug("GridManager2D_Free(): Null-Pointer !" )
  EndIf 
  ProcedureReturn *mem
EndProcedure  



; -----------------------------------------------------------------------------
; - bBoolean =  GridManager2D_Register(*grid.sGrid, *go.sGridObject2D)
; -----------------------------------------------------------------------------
; - Description:
; - Registers / Reregisters a Object in the Grid. It would be removed from
; - the Zones - if it was already registered and in the zones.
; - then it would be 'linked' to the Zones based on the postion and size
; -
; - Attention - you should not use this function directly. The 
; - Grid2DManager_Update() Reregisters all moved objects and Grid2DManager_AddObject() 
; - Makes the 'first' registration.
; -----------------------------------------------------------------------------
; - Returns:
; - Nothing
; -----------------------------------------------------------------------------

Procedure GridManager2D_Register(*grid.sGrid, *go.sGridObject2D)
   
  *based.sZonePtr = #Null 
  
  ;find Object in our registration
  *reg_obj.sGridObjPtr = FirstElement2(*grid\llRegisteredObjects) 
  While *reg_obj  
    If *reg_obj\obj=*go
      *based.sZonePtr = FirstElement2(*reg_obj\llZones) 
      Break 
    EndIf
    *reg_obj = NextElement2(*reg_obj) 
  Wend   
  
  ;If no Object there, we have to 'register' this one
  If *reg_obj = #Null 
    ;Remmeber registered Objects !!!
    *reg_obj.sGridObjPtr = AddElement2(*grid\llRegisteredObjects)
    *reg_obj\obj = *go 
    *reg_obj\llZones = NewList2(sZonePtr)
  EndIf
  
  ;remove from the zones where it was registered (if)
  If *based <> #Null
;;;    *based.sZonePtr = FirstElement2(*go\llZones) 
    While *based  
      *zone.sGridObjects = *based\zone 
      *obj.sGridObjPtr = FirstElement2(*zone\llGridObj) 
      While *obj 
        If *obj\obj = *go  
          DeleteElement2p(*zone\llGridObj,*obj)
          Break;
        EndIf
        *obj = NextElement2(*obj) 
      Wend 
      *based = NextElement2(*based) 
    Wend
    ;now delete the List on the Object
    ClearList2(*reg_obj\llZones)  
    LL_Free(*reg_obj\llZones)
    ;and create a new one.  (Bug in the LinkedList2...)
    *reg_obj\llZones  = NewList2(sZonePtr)
  Else    
  EndIf
  
   
  ;some precalculations
  sz = SizeOf(sGridObjects)
  wz = *grid\dimension\cx*sz
  
  ;Get the position and size of the object
  GridObject2D_GetBounds(*go, @object_rect.sRectL)
   
  ;Based on the Rect we check also the neighbour zones 
  px.l = (object_rect\X / *grid\gridElementSize\cx) - 1
  py.l = (object_rect\Y / *grid\gridElementSize\cy) - 1
  
  ;32 er grid ; object ist 50
  ; 50/32 = 1 + 1 = 2
  ; 90/32 = 2 + 1 = 3
  pw.l = px+(object_rect\width / *grid\gridElementSize\cx)+1+1
  ph.l = py+(object_rect\height / *grid\gridElementSize\cy)+1+1
  
  If pw > *grid\dimension\cx-1 : pw = *grid\dimension\cx-1 : EndIf
  If ph > *grid\dimension\cy-1 : ph = *grid\dimension\cy-1 : EndIf
  If px < 0 : px=0: EndIf
  If py < 0 : py=0: EndIf
  
  ;only serach in the 'defined' area istead throgh all zones
  For X = px To pw
    For Y = py To ph
      *zone.sGridObjects = *grid\paZones + (( Y*wz)+(X*sz))
      
      ;check collission 
      mcRectCollWrapper(*zone\rect, object_rect, Coll.l)
      If Coll
        ;Yes we are in the Zone
        *objRef.sGridObjPtr = AddElement2(*zone\llGridObj)
        *objRef\obj = *go
        
        ;and stor also the zone in the Object (what is not very 'clean' ! )
        *ZoneRef.sZonePtr = AddElement2(*reg_obj\llZones)
        *ZoneRef\zone = *zone 
      EndIf 
    Next 
  Next 
EndProcedure 
  

; -----------------------------------------------------------------------------
; - GridManager2D_AddObject(*grid.sGrid, *go.sGridObject2D) 
; -----------------------------------------------------------------------------
; - Description:
; - Add a Object to the Grid. It Registers the Object with the Grid and in the
; - Zones where it is.
; -----------------------------------------------------------------------------
; - Returns:
; - Nothing
; -----------------------------------------------------------------------------

Procedure GridManager2D_AddObject(*grid.sGrid, *go.sGridObject2D) 

  GridManager2D_Register(*grid,*go)
EndProcedure

; -----------------------------------------------------------------------------
; - bBoolean.l = GridManager2D_RemoveObject(*grid.sGrid, *go.sGridObject2D) 
; -----------------------------------------------------------------------------
; - Description:
; - Removes a Object from the Grid.
; - You have to do this, before you destroy that Object !
; -----------------------------------------------------------------------------
; - Returns:
; - #True  - if verything fine, otherwise #False (could not fond object)
; -----------------------------------------------------------------------------

Procedure GridManager2D_RemoveObject(*grid.sGrid, *go.sGridObject2D)  
  *based.sZonePtr = #Null 
  
  ;find Object in our registration
  *reg_obj.sGridObjPtr = FirstElement2(*grid\llRegisteredObjects) 
  While *reg_obj  
    If *reg_obj\obj=*go
      *based.sZonePtr = FirstElement2(*reg_obj\llZones) 
      Break 
    EndIf
    *reg_obj = NextElement2(*reg_obj) 
  Wend   
    
  If *reg_obj = #Null 
    ProcedureReturn #False ; but this is (maybe) wrong - you came up with some unregisterd
  EndIf
    
    ;remove from the zones where it was registered (if)
  If *based <> #Null
    ;;;    *based.sZonePtr = FirstElement2(*go\llZones) 
    While *based  
      *zone.sGridObjects = *based\zone 
      *obj.sGridObjPtr = FirstElement2(*zone\llGridObj) 
      While *obj 
        If *obj\obj = *go  
          DeleteElement2p(*zone\llGridObj,*obj)
          Break;
        EndIf
        *obj = NextElement2(*obj) 
      Wend 
      *based = NextElement2(*based) 
    Wend
    ;now delete the List on the Object
    DeleteList2(*reg_obj\llZones)  
  Else    
  EndIf
  ;and remove from the Registration !
  DeleteElement2p(*grid\llRegisteredObjects,*reg_obj)
  
  ProcedureReturn #True
EndProcedure
  
; -----------------------------------------------------------------------------
; - GridManager2D_Update(*grid.sGrid)
; -----------------------------------------------------------------------------
; - Description:
; - Goes through all objects that registerd with the Grid (Add Object !) and
; - checks them about changing the position (flag : HasMoved)
; - 
; - To make that work, you have to make sure to set the Flag after each movement.
;
; - alternativly (but maybe slower) - the Manager can Ask - (via the Update()-
; - Proc. at the Object.) Thats not here implemeted, but could added easily
; -----------------------------------------------------------------------------
; - Returns:
; - Nothing
; -----------------------------------------------------------------------------

Procedure GridManager2D_Update(*grid.sGrid)
  *obj.sGridObjPtr = FirstElement2(*grid\llRegisteredObjects) 
  While *obj  
    *current.sGridObject2D = *obj\obj
    If *current\HasMoved 
      GridManager2D_Register(*grid,*current)
    EndIf
    *obj = NextElement2(*obj) 
  Wend   
EndProcedure

; -----------------------------------------------------------------------------
; - llPointer = GridManager2D_GetObjectsInMyZones(*grid.sGrid, *go.sGridObject2D) 
; -----------------------------------------------------------------------------
; - Description:
; - Thats an Important Procedure - this gives you a Linked List of all
; - Objects fopund the same zones as the given Object. That are the one
; - you have to check about collission ! 
; - 
; - on the downside - you have to know how to handle the LinkedList2() and 
; - !!important !! you have to free the List after using it ! (DeleteList2())
; -----------------------------------------------------------------------------
; - Returns:
; - Pointer to LinkedList2 with Elements from Struct sGridObjPtr. 
; -----------------------------------------------------------------------------

Procedure.l GridManager2D_GetObjectsInMyZones(*grid.sGrid, *go.sGridObject2D) 
  *objList.sGridObjPtr = NewList2(sGridObjPtr)
  
  *zones.sZonePtr = #Null 
  
  ;find Object in our registration
  *reg_obj.sGridObjPtr = FirstElement2(*grid\llRegisteredObjects) 
  While *reg_obj  
    If *reg_obj\obj=*go
      *zones.sZonePtr = FirstElement2(*reg_obj\llZones) 
      Break 
    EndIf
    *reg_obj = NextElement2(*reg_obj) 
  Wend   
  
  If Not *zones = #Null   
  ;  *zones.sZonePtr = FirstElement2(*go\llZones) 
    While *zones  
      *zone.sGridObjects = *zones\zone 
      *obj.sGridObjPtr = FirstElement2(*zone\llGridObj) 
      While *obj 
        If *obj\obj <> *go  
          
          addAllowed.l = #True
          ;Already there ?
          *objX.sGridObjPtr = FirstElement2(*objList) 
          While *objX
            If *objX\obj = *obj\obj 
              addAllowed=#False
              Break;
            EndIf
            
            *objX = NextElement2(*objX) 
          Wend 
          
          If addAllowed
            *Otherobj.sGridObjPtr = AddElement2(*objList)
            *Otherobj\obj = *obj\obj
          EndIf
        EndIf
        *obj = NextElement2(*obj) 
      Wend 
      *zones = NextElement2(*zones) 
    Wend  
  EndIf 
  ProcedureReturn *objList
EndProcedure
  

; -----------------------------------------------------------------------------
; - GridManager2D_DrawZones(*go.sGridObject2D) 
; -----------------------------------------------------------------------------
; - Description:
; - Only to use for testing/Emaple - draws all Zones where the Object is
; - registerd. Since it is bsaicly a 'Box()'-Function you have do the
; - StartDrawing()/StopDrawing() around the call of tis Procedure
; -----------------------------------------------------------------------------
; - Returns:
; - Nothing
; -----------------------------------------------------------------------------
Procedure.l GridManager2D_DrawZones(*grid.sGrid, *go.sGridObject2D) 
  ;find Object in our registration
  *reg_obj.sGridObjPtr = FirstElement2(*grid\llRegisteredObjects) 
  While *reg_obj  
    If *reg_obj\obj=*go
      *zones.sZonePtr = FirstElement2(*reg_obj\llZones) 
      Break 
    EndIf
    *reg_obj = NextElement2(*reg_obj) 
  Wend   
  
  *zones.sZonePtr = FirstElement2(*reg_obj\llZones) 
  While *zones  
    *zone.sGridObjects = *zones\zone
    mcDrawRect(*zone\rect)
    *zones = NextElement2(*zones) 
  Wend   
EndProcedure
    

; -----------------------------------------------------------------------------
; - GridManager2D_DrawZonesDebug(*grid.sGrid) 
; -----------------------------------------------------------------------------
; - Description:
; - Only to use for testing/Emaple - draws all Zones in the Grid. The empty-ones
; - are green, where Objects regsiterd are red. 
; - StartDrawing()/StopDrawing() around the call of tis Procedure
; -----------------------------------------------------------------------------
; - Returns:
; - Nothing
; -----------------------------------------------------------------------------

Procedure.l GridManager2D_DrawZonesDebug(*grid.sGrid) 
  sz = SizeOf(sGridObjects)
  wz = *grid\dimension\cx*sz
  
  For X = 0 To *grid\dimension\cx-1
    For Y = 0 To *grid\dimension\cy-1
      *zone.sGridObjects = *grid\paZones + (( Y*wz)+(X*sz))
      If CountList2(*zone\llGridObj) > 0
        mcDrawRect(*zone\rect,#Red)
      Else
        mcDrawRect(*zone\rect,#Green) 
      EndIf 
    Next
  Next
EndProcedure


; jaPBe Version=3.8.2.693
; Build=0
; FirstLine=18
; CursorPosition=147
; ExecutableFormat=Windows
; DontSaveDeclare
; IDE Options = PureBasic 4.20 Beta 3 (Windows - x86)
; CursorPosition = 34
; FirstLine = 24
; Folding = ---