; GGR (Geographic Grid Resistration) 2D
; 
;  Here ist the Basic-Implementation of the GridObject2D
;  you can add your own properties to the sGridObject2D
;
; The Main-Idea is to register objects to one or more Zones on
; a Grid. So you can speed up you colission-checks - only for
; Objects on the 'Zones' where your Object is registered.
; imagine - you have not to check all of your Objects about
; a possible colission.
;
; I got the Idea from a XNA-Page (http://xna-uk.net/blogs/randomchaos/archive/2008/03/09/geographic-grid-registration-an-alternative-to-quadtrees-and-octrees.aspx)
; and he got it from the 'Game Programming Gems 6'-Book.... so here we go in PureBasic.... ;-)
;
; well we need to make huge usage of a dynamic LinkedList (from flype - thanks !)
; Also i added some Code-Fragments for Collision (by Kaeru Gaman) and this
; nice Procedure DisplayZoomRotateSprite3D() - but i dont know the author.
;
; So have a go - i will add this for example to my GameWizz - to Soeed up Game-Programming
; 
; one idea for future is a 3D-Version. but not needed at the moment.
;


;------- Includes ---
IncludeFile "_basic_structs.pbi"
IncludeFile "_basic_macros.pbi"
IncludeFile "_basic_procedures.pbi"

;------- Constants ---
;


;------- Macros ---


;------- Structures ---


; -----------------------------------------------------------------------------
; - This defines a Grid2D-Object
; - 
; - if you need more information/Data for your Objects, you can add Fields
; -----------------------------------------------------------------------------



Structure sGridObject2D
  id.l                ;Unique ID
  name.s              ;A Name for the Object - for Debug-Relations
  
  Size.SIZE           ;size 
  Scale.sPointF       ;Scale
  
  pos.sPointF         ; positione
  lastPos.sPointF     ;last-position
  
  HasMoved.l          ;
  
  boxColor.l          ;color for bounding box

  origin.sPointF
  
  Sprite.sSprite      
  
  angle.f
  
  Alpha.l
   
EndStructure






;Now are Procedure based on the Structure above


; -----------------------------------------------------------------------------
; - lPtrToObject =  GridObject2D_Create(name.s,id.l = -1)
; -----------------------------------------------------------------------------
; - Description:
; - Creates a new Grid2D-Object. 
; - At the Moment it Allocates the Memory for the sGridObject2D and fills
; - the struct with default-values.
; - to change these Values, you have to canged it on the returned pointer
; - to structure.
; -----------------------------------------------------------------------------
; - Returns:
; - a Pointer to a sGridObject2D if suceeds - otherwise #Null
; -----------------------------------------------------------------------------
Procedure.l GridObject2D_Create(name.s,id.l = -1)
  *mem.sGridObject2D = AllocateMemory(SizeOf(sGridObject2D))
  If *mem
    *mem\name = name
    *mem\id = id
    *mem\boxColor = #Red
    *mem\Size\cx = 50
    *mem\Size\cy = 50
    *mem\Scale\x = 1.0
    *mem\Scale\y = 1.0
    *mem\Alpha = 127
    *mem\angle = 270.0
  Else
    Debug("GridObject2D_Create("+name+"): Can't get Memory for Structure : sGridObject2D" )
  EndIf
  
  ProcedureReturn *mem
  
EndProcedure


; -----------------------------------------------------------------------------
; - GridObject2D_Update(*go.sGridObject2D, time.l)
; -----------------------------------------------------------------------------
; - Description:
; - Updates a new Grid2D-Object. 
; - At the Moment it  checks only position to trigger the hasMoved-Flag
; -----------------------------------------------------------------------------
; - Returns:
; - nothing
; -----------------------------------------------------------------------------

Procedure GridObject2D_Update(*go.sGridObject2D, time.l)
  If *go<> #Null
    ;Determine Movings
    If *go\pos\x <> *go\lastPos\x Or *go\pos\y <> *go\lastPos\y
      
      ;Copy new position as last known
      *go\lastPos\y = *go\pos\y
      *go\lastPos\x = *go\pos\x
      
      *go\HasMoved = #True
    Else
      *go\HasMoved = #False
    EndIf  
    
    ;Maybe here a place for your own additions
  
  Else
    Debug("GridObject2D_Update("+Str(*go)+", "+Str(time)+"): Null-Pointer !" )
  EndIf
EndProcedure


; -----------------------------------------------------------------------------
; - GridObject2D_GetBounds(*go.sGridObject2D, *rect.sRectL)
; -----------------------------------------------------------------------------
; - Description:
; - fills the given sRectL with the bounds of the Object.
; - also scale is used. So it should be the correct bounds.
; -----------------------------------------------------------------------------
; - Returns:
; - #True if it could fill the Rect-Scruture - otherwise #False
; -----------------------------------------------------------------------------

Procedure GridObject2D_GetBounds(*go.sGridObject2D, *rect.sRectL)
  If *go <> #Null And *rect <> #Null
    *rect\x =*go\pos\x
    *rect\y =*go\pos\y
    *rect\width  = (*go\Size\cx * *go\Scale\x)
    *rect\height = (*go\Size\cy * *go\Scale\y)
  Else
    Debug("GridObject2D_GetBounds("+Str(*go)+", "+Str(*rect)+"): Null-Pointer !" )
    ProcedureReturn #False
  EndIf
  ProcedureReturn #True
EndProcedure


; -----------------------------------------------------------------------------
; - bBoolean = GridObject2D_Load(*go.sGridObject2D, filename.s)
; -----------------------------------------------------------------------------
; - Description:
; - Loads a Sprite as Sprite3D and assign to the Object.
; - 
; - This is also one point to modificate for your needs !
; -----------------------------------------------------------------------------
; - Returns:
; - #True if it could Load and create the Sprite - otherwise #False
; -----------------------------------------------------------------------------

Procedure.l GridObject2D_Load(*go.sGridObject2D, filename.s)
  ;In the Default-State we use Sprite3D-Objects here
  If *go<> #Null
    *go\Sprite\id = LoadSprite(#PB_Any, filename ,#PB_Sprite_Texture | #PB_Sprite_AlphaBlending )
    If *go\Sprite\id <> 0
      *go\Sprite\s3d_id = CreateSprite3D(#PB_Any,*go\Sprite\id)   
    Else
      Debug("GridObject2D_Load("+Str(*go)+", "+filename+"): Can't load File !" )
      ProcedureReturn #False
    EndIf
  Else
    Debug("GridObject2D_Load("+Str(*go)+", "+filename+"): Null-Pointer !" )
    ProcedureReturn #False
  EndIf
  ProcedureReturn #True 
EndProcedure
  

; -----------------------------------------------------------------------------
; - bBoolean = GridObject2D_DrawBounds(*go.sGridObject2D)
; -----------------------------------------------------------------------------
; - Description:
; - Draws the Bounds of the Object
; - You have use this between 'StartDraing()/StopDrawing() - it uses a simple
; - Box() for drawing
; - 
; -----------------------------------------------------------------------------
; - Returns:
; - #True if it could Load and create the Sprite - otherwise #False
; -----------------------------------------------------------------------------

Procedure GridObject2D_DrawBounds(*go.sGridObject2D)
  If *go<> #Null 
    ;we assume thet StartDrawing() was done before !!
    GridObject2D_GetBounds(*go,@rect.sRectL)
    DrawingMode(#PB_2DDrawing_Outlined)
    Box(rect\x,rect\y,rect\width,rect\height, *go\boxColor)
  Else
    Debug("GridObject2D_DrawBounds("+Str(*go)+"): Null-Pointer !" )
    ProcedureReturn #False
  EndIf
  ProcedureReturn #True 
EndProcedure


; -----------------------------------------------------------------------------
; - bBoolean = GridObject2D_Draw(*go.sGridObject2D)
; -----------------------------------------------------------------------------
; - Description:
; - Draws / Diplays the Object.
; - We Display a Sprite3D per default, but you are free to chanmge this, to 
; - your own Displaying-code
; - 
; -----------------------------------------------------------------------------
; - Returns:
; - #True if it could Load and create the Sprite - otherwise #False
; -----------------------------------------------------------------------------

Procedure GridObject2D_Draw(*go.sGridObject2D)
  ;we assume thet Start3D() was done before !!
  If *go<> #Null  
    DisplayZoomRotateSprite3D(*go\Sprite\s3d_id,*go\pos\x,*go\pos\y,  (*go\Size\cx**go\Scale\x), (*go\Size\cy**go\Scale\y),*go\angle,0,0,*go\Alpha)
  Else
    Debug("GridObject2D_Draw("+Str(*go)+"): Null-Pointer !" )
    ProcedureReturn #False
  EndIf
  ProcedureReturn #True 
  
  
EndProcedure

  

; -----------------------------------------------------------------------------
; - bBoolean.l = GridObject2D_Destroy(*go.sGridObject2D)
; -----------------------------------------------------------------------------
; - Description:
; - Destroys a Objoect.
; - Also Loaded Sprites/Sprite3D are freed
; - 
; - You have to make Sure that this Object is unregistered 
; - from the Grid before ! 
; -----------------------------------------------------------------------------
; - Returns:
; - #True if it Destroyed - otherwise #False
; -----------------------------------------------------------------------------

Procedure.l  GridObject2D_Destroy(*go.sGridObject2D)
  Protected bRetVal.l = #True
  If Not *go = #Null
    If IsSprite3D(*go\Sprite\s3d_id) : FreeSprite3D(*go\Sprite\s3d_id) : EndIf
    If IsSprite(*go\Sprite\id)       : FreeSprite(*go\Sprite\id)       : EndIf 
    FreeMemory(*go)
  Else
    Debug("GridObject2D_Destroy("+mcDHex(*go)+"): Null Pointer !" )
    bRetVal = #False
  EndIf
  ProcedureReturn bRetVal 
EndProcedure
  


; jaPBe Version=3.8.2.693
; Build=0
; Language=0x0000 Language Neutral
; FirstLine=0
; CursorPosition=28
; ExecutableFormat=Windows
; DontSaveDeclare
; EOF