;---------------------------------------
; Object:  LinkedList2
; Version: Version 0.2
; Author:  Flype (flype44@hotmail.fr)
; Date:    Mar 2006
; Requir : Purebasic 4.0
;---------------------------------------


; MiniSet.pbi

Global NewList set.l()

Procedure.l Push(Value.l)
  If AddElement(set())
    set() = Value
  EndIf
EndProcedure
Procedure.l Pop(Value.l)
  ForEach set()
    If set() = Value
      DeleteElement(set())
    EndIf
  Next
EndProcedure
Procedure.l NotIn(Value.l)
  ForEach set()
    If set() = Value
      ProcedureReturn #False
    EndIf
  Next
EndProcedure

Structure sInOut
  mem.l
  Size.l
  line.l
  file.s 
EndStructure

Global NewList llIO.sInOut()

Macro DumpLLMemory()
  ForEach llIO()
    Debug("From File:"+llIO()\file +" line:"+Str(llIO()\line))
  Next
EndMacro



; LinkedList2.pbi

Structure LL
  *nextLL
  *prevLL
;  *parentLL   ;
  nByte.l
  nElement.l
  userdata.l
EndStructure

Macro NewTree(STRUCTNAME)
  LL_Alloc(SizeOf(STRUCTNAME)) : Debug("NewList2():"+#PB_Compiler_File+"(line:"+#PB_Compiler_Line+")") :; _Allocs()\info="NewList2():"+#PB_Compiler_File+"(line:"+#PB_Compiler_Line+")"
EndMacro

;  LL_Alloc(#PB_Compiler_File, #PB_Compiler_Line, SizeOf(STRUCTNAME))
Macro NewList2(STRUCTNAME)
  LL_Alloc(SizeOf(STRUCTNAME)) : Debug("NewList2():"+#PB_Compiler_File+"(line:"+Str(#PB_Compiler_Line)+")"):; _Allocs()\info="NewList2():"+#PB_Compiler_File+"(line:"+Str(#PB_Compiler_Line)+")"
EndMacro

Macro Assert2(element,func)
  CompilerIf #PB_Compiler_Debugger 
    If element = -1
      Debug "[" + func + "] List is empty"
    ElseIf element = -2
      Debug "[" + func + "] Out of bounds"
    ElseIf element = #Null
      Debug "[" + func + "] Pointer is null"
      CallDebugger
      End
    ElseIf NotIn(element)
      Debug "[" + func + "] Not a valid list"
      End
    EndIf
  CompilerEndIf
EndMacro

Procedure.l LL_Alloc( nByte.l=SizeOf(LL) )
  *list.LL = AllocateMemory(nByte)
  Push(*list)
  Assert2(*list,"NewList2")
  *list\nByte = nByte
  
  ; AddElement(llIO())
  ; llIO()\mem=*list
  ; llIO()\Size=nByte
  ; llIO()\file=file
  ; llIO()\line=line
  
  ProcedureReturn *list
EndProcedure
Procedure.l LL_Free(*element.LL)
  Assert2(*element,"LL_Free")
  ZeroMemory_(*element,*element\nByte)
  FreeMemory(*element)
  
  ; ForEach llIO()
    ; If llIO()\mem = *element
      ; DeleteElement(llIO())
      ; Break;
    ; EndIf
  ; Next  
  
  Pop(*element)
EndProcedure
Procedure.l LL_Link(*element1.LL,*element2.LL)
  Assert2(*element1,"LL_Attach")
  *element1\nextLL = *element2
  If *element2
    *element2\prevLL = *element1
  EndIf
EndProcedure

Procedure.l CountList2(*list.LL)
  Assert2(*list,"CountList2")
  If *list
    ProcedureReturn *list\nElement
  EndIf
EndProcedure
Procedure.l FirstElement2(*list.LL)
  Assert2(*list,"FirstElement2")
  If *list\nextLL
    ProcedureReturn *list\nextLL
  EndIf
  Assert2(-1,"FirstElement2")
EndProcedure
Procedure.l LastElement2(*list.LL)
  Assert2(*list,"LastElement2")
  While *list
    If *list\nextLL = #Null
      ProcedureReturn *list
    EndIf
    *list = *list\nextLL
  Wend
  Assert2(-1,"LastElement2")
EndProcedure
Procedure.l AddElement2(*list.LL,Position.l=-1)
  Assert2(*list,"AddElement2")
  *this.LL = LL_Alloc(*list\nByte)
  Assert2(*this,"AddElement2")
  *last.LL = LastElement2(*list)
  *list\nElement + 1
  If *last
    *this\prevLL = *last
    *last\nextLL = *this
  Else
    *this\prevLL = *list
  EndIf
  ProcedureReturn *this
EndProcedure
Procedure.l ClearList2(*list.LL)
  Assert2(*list,"ClearList2")
  *list\nElement = 0
  *list = FirstElement2(*list)
  While *list
    *this = *list
    *list = *list\nextLL
    LL_Free(*this)
  Wend
EndProcedure
Procedure.l DeleteList2(*list.LL)
  Assert2(*list,"DeleteList2")
  ClearList2(*list)
  LL_Free(*list)
EndProcedure
Procedure.l DeleteElement2(*list.LL,Position.l)
  Assert2(*list,"DeleteElement2")  
  *this.LL = FirstElement2(*list)
  While *this
    If i = Position
      LL_Link(*this\prevLL,*this\nextLL)
      LL_Free(*this)
      *list\nElement - 1
      ProcedureReturn
    EndIf
    *this = *this\nextLL
    i + 1
  Wend
  Assert2(-2,"DeleteElement2")
EndProcedure


Procedure.l DeleteElement2p(*list.LL,*this.LL)
  Assert2(*list,"DeleteElement2p")  
  Assert2(*this,"DeleteElement2p")  
  LL_Link(*this\prevLL,*this\nextLL)
  LL_Free(*this)
  *list\nElement - 1
  ProcedureReturn
EndProcedure

Procedure.l NextElement2(*list.LL)
  Assert2(*list,"NextElement2")
  ProcedureReturn *list\nextLL
EndProcedure
Procedure.l PreviousElement2(*list.LL)
  Assert2(*list,"PreviousElement2")
  *prev.LL = *list\prevLL
  If *prev
    If *prev\prevLL = #Null
      ProcedureReturn #Null
    EndIf
  EndIf
  ProcedureReturn *prev
EndProcedure
Procedure.l SwapElement2(*element1.LL,*element2.LL)
  Assert2(*element1,"SwapElement2")
  Assert2(*element2,"SwapElement2")
  *prev1 = *element1\prevLL
  *next1 = *element1\nextLL
  *element1\nextLL = *element2\prevLL
  *element1\nextLL = *element2\nextLL
  *element2\prevLL = *prev1
  *element2\nextLL = *next1
EndProcedure
Procedure.l SelectElement2(*list.LL,Position.l)
  Assert2(*list,"SelectElement2")
  *list = FirstElement2(*list)
  While *list
    If i = Position
      ProcedureReturn *list
    EndIf
    *list = *list\nextLL
    i + 1
  Wend
  Assert2(-2,"SelectElement2")
EndProcedure


Procedure.l InsertAfterElement2(*list.LL,*last.LL)
  Assert2(*list,"AddElement2")
  *this.LL = LL_Alloc(*list\nByte)
  Assert2(*this,"AddElement2")
  *list\nElement + 1
  If *last
    If *last\nextLL
      *this\nextLL = *last\nextLL
      *tNext.LL = *last\nextLL      
      *tNext\prevLL = *this
    EndIf    
    *this\prevLL = *last
    *last\nextLL = *this
  Else
    *this\prevLL = *list
  EndIf
  ProcedureReturn *this
EndProcedure


Procedure.l MoveElement2(*list.LL,*last.LL,*element.LL)
  Assert2(*list,"MoveElement2-1")
  Assert2(*last,"MoveElement2-2")
  Assert2(*element,"MoveElement2-3")  
  If *element\prevLL
    If *last\nextLL
      *element\nextLL = *last\nextLL
      *tNext.LL = *last\nextLL      
      *tNext\prevLL = *element
    EndIf    
  EndIf
  If *last  
    If *last\nextLL
      *element\nextLL = *last\nextLL
      *tNext.LL = *last\nextLL      
      *tNext\prevLL = *element
    EndIf    
    *element\prevLL = *last
    *last\nextLL = *element
  Else
    *element\prevLL = *last
  EndIf
  ProcedureReturn *element
EndProcedure



;--------------------------------------- 
; IDE Options = PureBasic v4.00 (Windows - x86)
; CursorPosition = 92
; FirstLine = 56
; jaPBe Version=3.8.2.693
; Build=0
; FirstLine=57
; CursorPosition=62
; ExecutableFormat=Windows
; DontSaveDeclare
; EOF